/*!
 * \file 
 * \brief Implementation of general functions to allocate memory. 
 * $Id: alloc.c,v 1.5 2005/04/26 18:10:34 hakonsen Exp $ 
 * 
 */
#include <stdio.h>
#include <malloc.h>
#include <stdlib.h>
#include "alloc.h"

/**************************************************************************
* Function for dynamic memory allocation of 1-D float array                
*--------------------------------------------------------------------------
* Input:  n1      - size of array                                          
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* First version:             Helge Coward        1999-09-29 
***************************************************************************/
 
float *matrix1_f(int n1)
{
  float *m;
  
  if ((m = (float *) malloc(n1*sizeof(float)))==NULL){
    fprintf(stderr,"Allocation error in matrix1_f\n");
    exit(1);
  }
  return m;
}
/**************************************************************************
* Function for dynamic memory allocation of 1-D float array with zero initial values               
*--------------------------------------------------------------------------
* Input:  n1      - size of array                                          
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* First version:      Greg Haakonsen        2005-04-21 
***************************************************************************/

float *matrix1_fz(int n1)
{
	float *m;
	if ((m = (float *) calloc(n1,sizeof(float)))==NULL)
	{
		fprintf(stderr,"Allocation error in matrix1_fz\n");
		exit(1);
	}
	return m;
}


/**************************************************************************
* Function for dynamic memory allocation of 2-D float array                
*--------------------------------------------------------------------------
* Input:  n1      - size of 1st dimension
*         n2      - size of 2nd dimension
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* First version:             Helge Coward        1999-09-29            
***************************************************************************/
 
float **matrix2_f(int n1,int n2)
{
  int i;
  float **m;

  if ((m = (float **) malloc(n1*sizeof(float*)))==NULL){
	 fprintf(stderr,"Allocation error in matrix2_f\n");
	 exit(1);
  }
  for (i=0;i<n1;i++){
	 if ((m[i]=(float *) malloc(n2*sizeof(float)))==NULL){
		fprintf(stderr,"Allocation error in matrix2_f\n");
		exit(1);
	 }
  }
  return m;
}

/**************************************************************************
* Function for dynamic memory allocation of 2-D float array with zero initial values              
*--------------------------------------------------------------------------
* Input:  n1      - size of 1st dimension
*         n2      - size of 2nd dimension
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* First version:				Greg Haakonsen		Feb 2005.         
***************************************************************************/
 
float **matrix2_fz(int n1,int n2)
{
  int i;
  float **m;

  if ((m = (float **) calloc(n1,sizeof(float*)))==NULL){
	 fprintf(stderr,"Allocation error in matrix2_f\n");
	 exit(1);
  }
  for (i=0;i<n1;i++){
	 if ((m[i]=(float *) calloc(n2,sizeof(float)))==NULL){
		fprintf(stderr,"Allocation error in matrix2_f\n");
		exit(1);
	 }
  }
  return m;
}

/**************************************************************************
* Function for dynamic memory allocation of 1-D double float array
*--------------------------------------------------------------------------
* Input:  n1      - size of array                                          
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* Last version:             Helge Coward        2001-03-01            
***************************************************************************/
 
double *matrix1_d(int n1)
{
  double *m;
  
  if ((m = (double *) malloc(n1*sizeof(double)))==NULL){
    fprintf(stderr,"Allocation error in matrix1_d\n");
    exit(1);
  }
  return m;
}
 
/**************************************************************************
* Function for dynamic memory allocation of 2-D double float array
*--------------------------------------------------------------------------
* Input:  n1      - size of 1st dimension
*         n2      - size of 2nd dimension
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* Last version:             Helge Coward        2001-03-01
***************************************************************************/
 
double **matrix2_d(int n1,int n2)
{
  int i;
  double **m;

  if ((m = (double **) malloc(n1*sizeof(double*)))==NULL){
	 fprintf(stderr,"Allocation error in matrix2_d\n");
	 exit(1);
  }
  for (i=0;i<n1;i++){
	 if ((m[i]=(double *) malloc(n2*sizeof(double)))==NULL){
		fprintf(stderr,"Allocation error in matrix2_d\n");
		exit(1);
	 }
  }
  return m;
}

/**************************************************************************
* Function for dynamic memory allocation of 1-D int array                
*--------------------------------------------------------------------------
* Input:  n1      - size of array                                          
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* Last version:             Helge Coward        1999-09-29            
***************************************************************************/
 
int *matrix1_i(int n1)
{
  int *m;
  
  if ((m = (int *) malloc(n1*sizeof(int)))==NULL){
    fprintf(stderr,"Allocation error in matrix1_i\n");
    exit(1);
  }
  return m;
}

/**************************************************************************
* Function for dynamic memory allocation of 1-D int array                
*--------------------------------------------------------------------------
* Input:  n1      - size of array                                          
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* First version:       Greg Haakonsen        2005-04-26            
***************************************************************************/
 
int *matrix1_iz(int n1)
{
	int *m;
	if ((m = (int *) calloc(n1, sizeof(int))) == NULL)
	{
		fprintf(stderr,"Allocation error in matrix1_iz\n");
		exit(1);
	}
	return m;
}
/**************************************************************************
* Function for dynamic memory allocation of 2-D int array                
*--------------------------------------------------------------------------
* Input:  n1      - size of 1st dimension
*         n2      - size of 2nd dimension
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* Last version:             Helge Coward        1999-09-29            
***************************************************************************/
 
int **matrix2_i(int n1,int n2)
{
  int i;
  int **m;

  if ((m = (int **) malloc(n1*sizeof(int*)))==NULL){
	 fprintf(stderr,"Allocation error in matrix2_i\n");
	 exit(1);
  }
  for (i=0;i<n1;i++){
	 if ((m[i]=(int *) malloc(n2*sizeof(int)))==NULL){
		fprintf(stderr,"Allocation error in matrix2_i\n");
		exit(1);
	 }
  }
  return m;
}

/**************************************************************************
* Function for dynamic memory allocation of 1-D char array                
*--------------------------------------------------------------------------
* Input:  n1      - size of array                                          
* Return: m       - pointer to array                                       
*--------------------------------------------------------------------------
* Last version:             Helge Coward        1999-09-29            
***************************************************************************/
 
char *matrix1_c(int n1)
{
  char *m;
  
  if ((m = (char *) malloc(n1*sizeof(char)))==NULL){
    fprintf(stderr,"Allocation error in matrix1_c\n");
    exit(1);
  }
  return m;
}

/**************************************************************************
* Function for freeing 2-D array                
*--------------------------------------------------------------------------
* Input: m       - pointer to array                                       
*        n1      - size of 1st dimension
*--------------------------------------------------------------------------
* First version:             Helge Coward        1999-09-30 
***************************************************************************/

void free_2(void **m, int n1)
{
  int i;
  for (i=0;i<n1;i++)
	 free(m[i]);
  free(m);
}
/**************************************************************************
* Function for freeing 2-D array               
*--------------------------------------------------------------------------
* Input	: pp		- pointer to pointer to array
* 		: n1		- Number of rows.
*--------------------------------------------------------------------------
* First version:     Greg Haakonsen        2005-04-26 
***************************************************************************/

void ** FreeVect2( void **pp, int n1 )
{
	int i;
	if (pp != NULL)
	{
		for (i = 0; i < n1; i++) 
		{
			if ( pp[i] != NULL )
			{
				free(pp[i]);
				pp[i] = NULL;
			}
		}
		free(pp);
		pp = NULL;
	}
	return (pp);
}

/**************************************************************************
* Function for freeing 1-D array               
*--------------------------------------------------------------------------
* Input: p       - pointer to array                                       
*--------------------------------------------------------------------------
* First version:     Greg Haakonsen        2005-04-26 
***************************************************************************/

void * FreeVect( void *p )
{
	if ( p != NULL )
	{
		free(p);
		p = NULL;
	}
	return(p);
}

unsigned char **matrix2_uch(int Rows, int Cols)
{
	int i ;
	unsigned char **m;

	m = (unsigned char **) malloc((unsigned) Rows*sizeof(unsigned char *));
	if (!m)
	{
		fprintf(stderr,"Allocation error in matrix2_uch\n");
		exit(1);
	}
	for (i=0;i<Rows;i++)
	{
	      m[i]=(unsigned char*) malloc((unsigned) Cols*sizeof(unsigned char));
		if (!m[i])
		{
			fprintf(stderr,"Allocation error in matrix2_uch\n");
			exit(1);
		}
	}
	return m;
}

unsigned char *matrix1_uch(int Length)
{
	unsigned char *m;

	m = (unsigned char *) malloc((unsigned) Length*sizeof(unsigned char));
	if (!m)
	{ 
		fprintf(stderr,"Allocation error in matrix1_uch\n");
		exit(1);
	}
	return m;
}

void freematrix2_uch(unsigned char **m,int rows,int cols)
{
	int i;
	for (i=rows-1;i>=0;i--){
		free((char*) (m[i]));
	}
	free((char*) (m));
}

void freematrix2_f(float **m,int rows,int cols)
{	
	int i;
	if (rows == 122880) {
		for (i=rows-1;i>=0;i--){		
//			if ((i%2000)==0) {
//				printf("i= %i\n",i);
//			}
			free((char*) (m[i]));
		}
	}
	else {
		for (i=rows-1;i>=0;i--){		
			free((char*) (m[i]));
		}
	}
	free((char*) (m));
}

void freematrix1_uch(unsigned char *m)
{ 
	free((char*) m);
}

