/*!
 * \file
 * \brief Declaration of functions for allocation of memory.
 * \author First written by Helge Coward and Arild Fuldseth et al.
 * \author Last version by Greg Haakonsen
 */
#ifndef _ALLOC_H
#define _ALLOC_H

/*!
 * \brief Allocate memory, n1 floats.
 * \author Helge Coward
 * \param[in] n1 number of floats in vector.
 * \return pointer to vector.
 */
extern float *matrix1_f(int n1);
/*!
 * \brief Allocate memory, n1 floats. Set to zero initially.
 * \author Greg Haakonsen
 * \param[in] n1 number of floats in vector.
 * \return pointer to vector.
 */
extern float *matrix1_fz(int n1);
/*!
 * \brief Allocate memory array, n1 by n2 floats.
 * \author Helge Coward
 * \param[in] n1 size of first dimension.
 * \param[in] n2 size of second dimension.
 * \return pointer to vector of pointers.
 */
extern float **matrix2_f(int n1,int n2);
/*!
 * \brief Allocate memory array, n1 by n2 floats. Initialize to zero.
 * \author Greg Haakonsen, based on Helge Cowards version.
 * \param[in] n1 size of first dimension.
 * \param[in] n2 size of second dimension.
 * \return pointer to vector of pointers.
 */
extern float **matrix2_fz(int n1,int n2);
/*!
 * \brief Allocate memory vector, n1 doubles.
 * \author Helge Coward
 * \param[in] n1 size of first dimension.
 * \return pointer to vector.
 */
extern double *matrix1_d(int n1);
/*!
 * \brief Allocate memory array, n1 by n2 doubles.
 * \author Helge Coward
 * \param[in] n1 size of first dimension.
 * \param[in] n2 size of second dimension.
 * \return pointer to vector of pointers.
 */
extern double **matrix2_d(int n1,int n2);
/*!
 * \brief Allocate memory vector, n1 integers.
 * \author Helge Coward
 * \param[in] n1 size of first dimension.
 * \return pointer to vector.
 */
extern int *matrix1_i(int n1);
/*!
 * \brief Allocate memory, n1 integers. Set to zero initially.
 * \author Greg Haakonsen, based on Helge Cowards version.
 * \param[in] n1 number of ints in vector.
 * \return pointer to vector.
 */
extern int *matrix1_iz(int n1);
/*!
 * \brief Allocate memory array, n1 by n2 integers.
 * \author Helge Coward
 * \param[in] n1 size of first dimension.
 * \param[in] n2 size of second dimension.
 * \return pointer to vector of pointers.
 */
extern int **matrix2_i(int n1,int n2);
/*!
 * \brief Allocate memory n1 characters.
 * \author Helge Coward
 * \param[in] n1 size of first dimension.
 * \return pointer to vector.
 */
extern char *matrix1_c(int n1);
/*!
 * \brief Freeing 2D array.
 * \author Helge Coward
 * \param[in] n1 size of first dimension.
 * \param[in] m pointer to pointers.
 */
extern void free_2(void **m,int n1);
/*!
 * \brief Free memory used for a general matrix. Checks if pointers are set to NULL.
 * \author Greg Haakonsen
 * \param[in] n1 size of first dimension.
 * \param[in] pp pointer to pointers.
 * \return Null Pointer.
 */
extern void ** FreeVect2( void **pp, int n1 );
/*!
 * \brief Free memory used for a general vector. Sets pointer to NULL. Cast returned value to whatever.
 * \author Greg Haakonsen
 * \param[in] p pointer to vector.
 * \return Null pointer.
 */
extern void * FreeVect( void *p );

void freematrix1_uch(unsigned char *m);
void freematrix2_f(float **m,int rows,int cols);
void freematrix2_uch(unsigned char **m,int rows,int cols);
unsigned char *matrix1_uch(int Length);
unsigned char **matrix2_uch(int Rows, int Cols);




#endif
