#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "constants.h"
#include "filtering_k.h" 
#include "alloc.h"
 
/***************************************************************************
* Function for 2D analysis parallel filterbank with decimation                 
*---------------------------------------------------------------------------
* Input:   nrows    - number of rows                         
*          ncols    - number of columns
*          x        - original frame 
*          fil_par  - filter data                      
*          qflag    - quiet flag  
* Output:  sub1     - subband frame
*---------------------------------------------------------------------------
* Last version:    Arild Fuldseth        1996-07-16 
***************************************************************************/

void Anl_fbank_Par(int nrows, int ncols, float *x,float *sub1,
						 struct firspec_par fil_par, int screenflag)
{
  int   j,v,row,col,nch,ntaps;
  float acc,*line_in,*line_out,*fptr1,*fptr2,*fptr3;
 
  if (screenflag)
	 fprintf(stderr,"Analysis filtering - parallel structure\n");

  nch = fil_par.channels;          /* no. of channels */
  ntaps = fil_par.no_of_taps;      /* no. of filter coef. */

  if (ncols > nrows){
    line_in = matrix1_f(ncols);
    line_out = matrix1_f(ncols+ntaps+1);
  }
  else{
    line_in = matrix1_f(nrows);
    line_out = matrix1_f(nrows+ntaps+1);
  }

  /* filtering horizontally */

  for (row=0;row<nrows;row++){          /* for each row */     
    for (col=0;col<ncols;col++){
		line_in[col] = x[row*ncols+col];
	 }
    mirror_anl(ncols,ntaps,nch,line_in,line_out);
    fptr3 = &sub1[row*ncols];
    for (j=0;j<nch;j++){                /* for each horizontal subband */
      for (col=0;col<ncols;col+=nch){   /* for each sample */
        acc = 0.0;
        fptr1 = &fil_par.a[j][0];
        fptr2 = &line_out[col];
        for (v=ntaps;v>0;v--){
			 acc += *fptr1++ * *fptr2++;
		  }
        *fptr3++ = acc;
      }
    }
  }


  /* filtering vertically */

  for (col=0;col<ncols;col++){          /* for each column */     
    for (row=0;row<nrows;row++){
		line_in[row] = sub1[row*ncols+col];
	 }
    mirror_anl(nrows,ntaps,nch,line_in,line_out);
    fptr3 = &sub1[col];
    for (j=0;j<nch;j++){                /* for each vertical subband */
      for (row=0;row<nrows;row+=nch){   /* for each sample */ 
		  acc = 0.0;
		  fptr1 = &fil_par.a[j][0];
		  fptr2 = &line_out[row];
		  for (v=ntaps;v>0;v--){
			 acc += *fptr1++ * *fptr2++;
		  }
		  *fptr3 = acc;
		  fptr3 += ncols;
      }
    }
  }
  
  /* free memory */
 
  free(line_in);
  free(line_out);
}

 
/***************************************************************************
* Function for 2D synthesis parallel filterbank of decimated input signal      
*--------------------------------------------------------------------------
* Input:   nrows   - number of rows                           
*          ncols   - number of columns                      
*          sub1    - quantized subband frame 
*          fil_par - filter data
*          qflag   - quiet flag  
* Output:  y       - reconstructed frame        
*--------------------------------------------------------------------------
* Last version:     Arild Fuldseth        1996-07-16            
***************************************************************************/

void Syn_fbank_Par(int nrows,int ncols,float *sub1,float *y,
						 struct firspec_par fil_par, int screenflag)
{
  int   j,m,row,col,nch,ntaps,nch2,index,offset,offset2;
  float acc,*coef,*fptr1,*fptr2,*fptr3,*line_in,*line_out;

  if (screenflag)
	 fprintf(stderr,"Synthesis filtering - parallel structure\n");

  /* initialize */
  
  nch = fil_par.channels;          /* no. of channels */
  ntaps = fil_par.no_of_taps;      /* no. of filter coef. */
  nch2 = nch*nch;
  offset = (ntaps-nch)/2;
  offset2 = ntaps/nch - (offset+nch)/nch;
 
  if (ncols > nrows){
    line_in = matrix1_f(ncols);
    line_out = matrix1_f(ncols+2*offset2*nch);
  }
  else{
    line_in = matrix1_f(nrows);
    line_out = matrix1_f(nrows+2*offset2*nch);
  }
  coef = matrix1_f(ntaps*nch);


  /* transpose coefficient matrix for fast implementation */

  for (m=0; m<ntaps; m++){
    for (j=0; j<nch; j++){ 
		coef[m*nch+j] = fil_par.b[j][m];
	 }
  } 


  /* vertical filtering */  

  for (col=0;col<ncols;col++){           /* for each column */
    for (row=0;row<nrows;row++){ 
		line_in[row] = sub1[row*ncols+col];
	 }
    mirror_syn(nrows,ntaps,nch,line_in,line_out);
    for (row=0;row<nrows;row++){         /* for each row */
      index = row+offset;
      acc = 0.0;
      fptr3 = &coef[(index%nch+ntaps-nch)*nch];
      fptr2 = &line_out[nch*(index/nch-offset/nch)];
      for (m=ntaps/nch; m>0; m--){       /* for each filter coefficient */
		  fptr1 = fptr3;
		  for (j=0;j<nch;j++){
			 acc += *fptr1++ * *fptr2++;
		  }
		  fptr3 -= nch2;
      }
      sub1[row*ncols+col] = acc;
    }
  }

  /* horizontal filtering */  

  for (row=0;row<nrows;row++){            /* for each row */
    for (col=0;col<ncols;col++){
		line_in[col] = sub1[row*ncols+col];
	 }
    mirror_syn(ncols,ntaps,nch,line_in,line_out);
    for (col=0;col<ncols;col++){          /* for each column */
      index = col+offset;
      acc = 0.0;
      fptr3 = &coef[(index%nch+ntaps-nch)*nch];
      fptr2 = &line_out[nch*(index/nch-offset/nch)];
      for (m=ntaps/nch; m>0; m--){        /* for each filter coefficient */
		  fptr1 = fptr3;
		  for (j=0;j<nch;j++){
			 acc += *fptr1++ * *fptr2++;
		  }
		  fptr3 -= nch2;
      }
      y[row*ncols+col] = acc;
    }
  }  
  
  /* free memury */
  
  free(line_in);
  free(line_out);
  free(coef);
}

/**************************************************************************
* Function to apply 1D parallel synthesis mirror imaging to a line of data     
*--------------------------------------------------------------------------
* Input:  len      - length of input data                             
*         ntaps    - number of filter coefficients
*         nch      - number of subbands                                        
*         line_in  - input data                                
* Output: line_out - output data                                
*--------------------------------------------------------------------------
* Last version:             Arild Fuldseth        1996-07-16            
***************************************************************************/

void mirror_syn(int len, int ntaps, int nch, float *line_in,
					 float *line_out)
{

  int j,m,off,off2,blen;

  blen = len/nch;          
  off = (ntaps-nch)/2;
  off2 = ntaps/nch - (off+nch)/nch;  

  /* define subband samples for m < 0 */

  for (m=-off2;m<0;m++){       
    for (j=0;j<nch;j+=2){
		line_out[nch*(m+off2)+j] = line_in[blen*j+m-2*m-1];
	 }
    for (j=1;j<nch;j+=2){ 
		line_out[nch*(m+off2)+j] = -line_in[blen*j+m-2*m-1];
	 }
  }
  
  /* define subband samples for 0 <= m < blen */
  
  for (m=0;m<blen;m++){      
    for (j=0;j<nch;j++){ 
		line_out[nch*(m+off2)+j] = line_in[blen*j+m];
	 }
  }
  
  /* define subband samples for m > 0 */
  
  for (m=blen;m<blen+off2;m++){  
    for (j=0;j<nch;j+=2){
		line_out[nch*(m+off2)+j] = line_in[blen*j+m-2*(m-blen)-1];
	 }
    for (j=1;j<nch;j+=2){ 
		line_out[nch*(m+off2)+j] =	-line_in[blen*j+m-2*(m-blen)-1];
	 }
  }
}

/**************************************************************************
* Function to apply 1D parallel analysis mirror imaging to a line of data      
*--------------------------------------------------------------------------
* Input:  len      - length of input data                             
*         ntaps    - number of filter coefficients
*         nch      - number of subbands                      
*         line_in  - input data                                
* Output: line_out - output data                                
*--------------------------------------------------------------------------
* Last version:             Arild Fuldseth        1996-07-16            
***************************************************************************/

void mirror_anl(int len,int ntaps,int nch,float *line_in,float *line_out){
 
  int n,offset;
  offset = (ntaps-nch)/2;

  for (n=-offset;n<0;n++){ 
	 line_out[n+offset] = line_in[-n-1];
  }
  for (n=0;n<len;n++){
	 line_out[n+offset] = line_in[n];
  }
  for (n=len;n<=len+(ntaps+nch)/2;n++){ 
	 line_out[n+offset] = line_in[2*len-n-1];
  }
}


/***************************************************************************
* Function for 2D analysis filterbank                                         
*---------------------------------------------------------------------------
* Input:   nrows            - number of rows                         
*          ncols            - number of columns
*          x                - original image 
*          fil_tree         - filter data
*          qflag            - quiet flag
* Output:  sub1             - subband image
*---------------------------------------------------------------------------
* First version:  Ilangko Balasingham   1996-10-10
***************************************************************************/

void Anl_fbank_Tree(int nrows, int ncols, float *x, float *sub1,
						  struct firspec_tree fil_tree, int screenflag)
{
  int   row,col,k,no,scale;
  float *line_in,*line_out;
  
  if (screenflag)
	 fprintf(stderr,"Analysis filtering - tree structure \n");

  if (ncols > nrows){
    line_in = matrix1_f(ncols);
    line_out = matrix1_f(ncols);
  }
  else{
    line_in = matrix1_f(nrows);
    line_out = matrix1_f(nrows);
  }
  
  /* Initialization */
   
  for (row=0;row<nrows;row++){
    for (col=0;col<ncols;col++){
      sub1[row*ncols+col] = x[row*ncols+col];
    }
  }
  
  /* analysis filtering */

  /* Analysis uniform filtering: tree-structured implementation */
  
  scale = NCH;
  for (k=0;k<(fil_tree).levels;k++){
	 /*fprintf(stderr,"  Filtering horizontally\n");*/
	 no=k;

	for (row=0;row<nrows/scale;row++){
	  for (col=0;col<ncols/scale;col++){
		 line_in[col] = sub1[row*ncols+col];
	  }
	  fir_n_filt_m_ana(no,ncols/scale,line_in,line_out,fil_tree);
	  for (col=0;col<ncols/scale;col++){
		 sub1[row*ncols+col] = line_out[col];
	  }
	}
	/*fprintf(stderr,"  Filtering vertically\n");*/
	for (col=0;col<ncols/scale;col++){
	  for (row=0;row<nrows/scale;row++){
		 line_in[row] = sub1[row*ncols+col];
	  }
	  fir_n_filt_m_ana(no,nrows/scale,line_in,line_out,fil_tree);
	  for (row=0;row<nrows/scale; row++){
		 sub1[row*ncols+col] = line_out[row];
	  }
	}
	scale*=2; 
 }
 
  /* free memory */
  
  free(line_in);
  free(line_out);
  
}


/***************************************************************************
* Function for 2D synthesis filterbank                                        
*--------------------------------------------------------------------------
* Input:   nrows                 - number of rows                           
*          ncols                 - number of columns                      
*          sub2                  - quantized subband image
*          fil_tree              - filter data 
*          qflag                 - quiet flag
* Output:  sub1                  - reconstructed T-image        
*--------------------------------------------------------------------------
* First version:   Ilangko Balasingham   1996-10-10
***************************************************************************/

void Syn_fbank_Tree(int nrows, int ncols, float *sub1, float *y,
						  struct firspec_tree fil_tree, int screenflag)
{
  int   row,col,k,no,scale,levels;
  float *line_in,*line_out;

  if (screenflag)
	 fprintf(stderr,"Synthesis filtering - tree structure \n");

  /* initialize */

  if (ncols > nrows){
    line_in = matrix1_f(ncols);
    line_out = matrix1_f(ncols);
  }
  else{
    line_in = matrix1_f(nrows);
    line_out = matrix1_f(nrows);
  }
  
  
  /* Initialization */
  
  for (row=0;row<nrows;row++){
    for (col=0;col<ncols;col++){
     y[row*ncols+col] = sub1[row*ncols+col];
    }
  }

  /* Synthesis uniform filtering: tree-structured implementation */
  
  /*scale=(int)pow(2,((fil_tree).levels-1))*8;*/ /* starting stage for
	 inverse filtering*/
  scale = 1 << (((fil_tree).levels-1) + NPARLEVELS);
  no=(fil_tree).levels-1;

  levels=(fil_tree).levels;

  for (k=0;k<levels;k++){
	 /* fprintf(stderr,"  Filtering vertically\n"); */
	 for (col=0;col<ncols/scale;col++){
		for (row=0;row<nrows/scale;row++){
		  line_in[row] = y[row*ncols+col];
		}
		fir_n_filt_m_syn(no,nrows/scale,line_in,line_out,fil_tree);
		for (row=0;row<nrows/scale; row++){
		  y[row*ncols+col] = line_out[row];
		}
	 }
	 /*  fprintf(stderr,"  Filtering horizontally\n"); */
	 for (row=0;row<nrows/scale;row++){
		for (col=0;col<ncols/scale;col++){
		  line_in[col] = y[row*ncols+col];
		}
		fir_n_filt_m_syn(no,ncols/scale,line_in,line_out,fil_tree); 
		for (col=0;col<ncols/scale;col++){
		  y[row*ncols+col] = line_out[col];
		}
	 }
	 no--;
	 scale=scale/2;
	 
  }
  
  free(line_in);
  free(line_out);
}



/***************************************************************************
* Function implementing 1D analysis filterbank using parallel FIR filters     
* and mirror extension for one line of input image data. 
* NB! New restrictions.........           
*---------------------------------------------------------------------------
* Input:   len                   - signal length                            
*          line_in               - one line of image data for filter input
*          no                    - filter stage number
*          fil_tree               - filter data
* Output   line_out              - one line of image data for filter output
*---------------------------------------------------------------------------
* First version:                John H. Husoy                              
* Second version:               Ilangko Balasingham   1996-10-10
***************************************************************************/

void fir_n_filt_m_ana(int no, int len, float *line_in, 
							 float *line_out, struct firspec_tree fil_tree)
{ 
  int p,n,m,j,offset,blen,ch,taps_lp,taps_hp,taps,subind,start;

  /* Initialize */
  
  ch = NOOFCHANNELS;          /* No. of channels */
  taps_lp = (fil_tree).nlptaps[no];      /* No. of filter coef. LP*/
  taps_hp = (fil_tree).nhptaps[no];      /* No. of filter coef. HP*/
  
  
#ifdef PARAM_CHECK
  /* Check parameters */
  /* If the program calling this is programmed correctly, the checks
	  are unnecessary */
  
  if (ch%2 != 0){
    fprintf(stderr,"Illegal no. of channels coefficients\n");
    exit(1);
  }
  if (len%ch != 0){
    fprintf(stderr,"Illegal image size\n");
    exit(1);
  }
#endif

  /* Prepare for filtering */

  blen = len/ch;              /* No. of samples in each subband */


  /* Check whether filter length is even or odd */
  
  if (taps_lp%2==0){   /* Even length filter bank */
    for (j=0;j<ch;j++) { 
      if (j==0){     /* LP filter */
		  taps=taps_lp;
		  /* p=(taps+ch)/2-1; */
			  /* offset=taps-1-p; */
		  offset=(taps-ch)/2;
      }
      else{         /* HP filter */
		  taps=taps_hp;
		  /* p=(taps+ch)/2-1; */
			  /* offset=taps-1-p; */
		  offset=(taps-ch)/2;
      }
      
      for (m=0;m<blen;m++){
		  subind=m+blen*j;
		  line_out[subind]=0.0;
		  start=m*ch-offset;
		  for (n=start;n<start+taps;n++){
			 if (n < 0){
				line_out[subind] += (fil_tree).ana[no][j][n-start]*line_in[-n-1];
			 }
			 else if (n < len){
				line_out[subind] += (fil_tree).ana[no][j][n-start]*line_in[n];
			 }
			 else{  /* n must be >= len */
				line_out[subind] += 
				  (fil_tree).ana[no][j][n-start]*line_in[2*len-n-1];
			 }
		  }
      }
    }
  }
  else{  /* Odd length filter bank */     
    
    for (j=0;j<ch;j++){     /* For each channel */
      if (j==0){
		  p=(taps_lp-1)/2;  /* Even symmetry */
		  taps=taps_lp;
		  offset=taps-1-p;  /* First signal value to be filtered */
      }
      else{
		  p=(taps_hp-1)/2+1; /* Odd symmetry */
		  taps=taps_hp;
		  offset=taps-1-p;  /* First signal value to be filtered */
      }
      
      for (m=0;m<blen;m++){                /* for each output sample */
		  subind = blen*j+m;
		  line_out[subind] = 0.0;
		  start=m*ch-offset;
		  for (n=start;n<start+taps;n++){ /* for each filter coef. */
			 if (n < 0){
				line_out[subind] +=
				  (fil_tree).ana[no][j][n-start]*line_in[-n];
			 }
			 else if (n < len){
				line_out[subind] +=
				  (fil_tree).ana[no][j][n-start]*line_in[n];
			 }
			 else{
				line_out[subind] +=
				  (fil_tree).ana[no][j][n-start]*line_in[2*len-n-2];
			 }
        }
      }
    }     
  }
}



/***************************************************************************
* Function implementing 1D synthesis filterbank using parallel FIR filters    
* and mirror extension for one line of input image data
* NB! New restrictions.....                    
*---------------------------------------------------------------------------
* Input:   len                   - signal length                         
*          line_in               - one line of image data for filter input
*          no                    - filter stage number
*          fil_tree               - filter data
* Output   line_out              - one line of image data for filter output
*---------------------------------------------------------------------------
* First version:                John H. Husoy                              
* Second version.               Ilangko Balasingham   1996-10-10
***************************************************************************/

void fir_n_filt_m_syn(int no, int len, float *line_in, 
							 float *line_out, struct firspec_tree fil_tree)
{
  int   n,m,j,r,offset,blen,ch,taps,subind,cofind;
  int   taps_lp, taps_hp, symmetype;
  
  
  /* Initialize */    
  ch = NOOFCHANNELS;                        /* no. of channels */
  taps_lp = (fil_tree).nhptaps[no];      /* no. of filter coef. LP*/
  taps_hp = (fil_tree).nlptaps[no];      /* no. of filter coef. HP*/
  
#ifdef PARAM_CHECK
  /* Check parameters */
  /* If the program calling this is programmed correctly, the checks
	  are unnecessary */

  if (ch%2 != 0){
    fprintf(stderr,"Illegal no. of channels coefficients\n");
    exit(1);
  }
  if (len%ch != 0){
    fprintf(stderr,"Illegal image size\n");
    exit(1);
  }
#endif
  
  /* prepare filtering */
  
  blen = len/ch;             /* no. of samples in each subband */
  for (j=0;j<len;j++){
    line_out[j] = 0.0;
  }
  
  
  /* Check whether filter length is even or odd */  
  
  if (taps_lp%2==0){ /* Even length filters */
    for (j=0;j<ch;j++){
      if (j%2==0){  /* LP filter */
		  taps=taps_lp;
		  r=(taps-ch)/2;
		  symmetype=1; /* Even symmetry */
      }
      else{        /* HP filter */
		  taps=taps_hp;
		  r=(taps-ch)/2;
		  symmetype=0; /* Odd symmetry */
      }
      for (n=0;n<len;n++){
		  offset=n+r;
		  for (m=(int)ceil((offset-(taps-1))/((float) ch));
				 m<=floor(offset/((float) ch));m++){
			 subind=blen*j+m;
			 cofind=offset-m*ch;
			 if (m < 0){
				if (symmetype==1) {/* Even symmetry */
				  line_out[n] += 
					 (fil_tree).syn[no][j][cofind]*line_in[subind-2*m-1];
				}     
				else { /* Odd symmetry */
				  line_out[n] -= 
					 (fil_tree).syn[no][j][cofind]*line_in[subind-2*m-1];
				}
			 }
			 else if (m < blen){
				line_out[n] += 
				  (fil_tree).syn[no][j][cofind]*line_in[subind];
			 }
			 else {  /* m must be >= blen */
				if (symmetype==1) { /* Even symmetry */
				  line_out[n] += 
					 (fil_tree).syn[no][j][cofind]*line_in[subind-2*(m-blen)-1];
				}
				else { /* Odd symmetry */
				  line_out[n] -= 
					 (fil_tree).syn[no][j][cofind]*line_in[subind-2*(m-blen)-1];
				}
			 }
		  }
      }
    }
  }
  else {   /* Odd length filter bank */
    for (j=0;j<ch;j++){            /* for each channel */
      if (j%2==0){              /* LP channel */
		  taps=taps_lp;
		  symmetype=1; /* Even symmetry */
		  r = (taps-1)/2;
      }
      else {
		  taps=taps_hp;
		  symmetype=0; /* Odd symmetry */
		  r=(taps-1)/2-1;
      }
      for (n=0;n<len;n++){ /* for each output sample */
		  offset = n+r;
		  for (m=(int)ceil((offset-(taps-1))/((float) ch));
				 m<=floor(offset/((float) ch));m++){   
			 subind=blen*j+m;
			 cofind=offset-m*ch;
			 if (m < 0){
				if (symmetype==1) { /* Even symmetry: Use WHOLE
											  sample symmetry */
				  line_out[n] += (fil_tree).syn[no][j][cofind]*
					 line_in[subind-2*m];
				} 
				else {             /* Odd symmetry: Use HALF
											 sample symmetry */
				  line_out[n] += (fil_tree).syn[no][j][cofind]*
					 line_in[subind-2*m-1];
				}         
				
			 }
			 else if (m < blen){
				line_out[n] += (fil_tree).syn[no][j][cofind]*
				  line_in[subind];
			 }
			 else{ 
				if (symmetype==1) { /* Even symmetry: Use HALF
											  sample symmetry */
				  line_out[n] += (fil_tree).syn[no][j][cofind]*
					 line_in[subind-2*(m-blen)-1];
				}
				else { /* Odd symmetry: Use WHOLE sample
							 symmetry */
				  line_out[n] += (fil_tree).syn[no][j][cofind]*
					 line_in[subind-2*(m-blen)-2];
				} 
			 }
		  }
      }
    }
  }         
}







