#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/file.h>
#include <string.h>

#include "constants.h"
#include "util.h"
#include "codecdecl.h"


/***************************************************************************
* Check image size and expand if necessarry   
*--------------------------------------------------------------------------
* Input:     nrows       - number of rows
*            ncols       - number of columns
*            v_blksz     - vertical blocksize
*            h_blksz     - horizontal blocksize
*            nparlevels  - number of parallel levels
*            ndylevels   - number of dyadic levels
* Output:    nrowse      - expanded number of rows
*            ncolse      - expanded number of columns
*--------------------------------------------------------------------------
* First version:          Helge Coward        2001-03-08          
***************************************************************************/
void expanded_size(int nrows, int ncols, int v_blksz, int h_blksz, 
						 int nparlevels, int ndylevels, int *nrowse, 
						 int *ncolse)
{
  int div, div_v, div_h, nparbands;
  
  nparbands = 1<<nparlevels;  
  div = 1<<(nparlevels+ndylevels);
  div_v = least_common_multiplier(div, nparbands*v_blksz);
  div_h = least_common_multiplier(div, nparbands*h_blksz);
	
  *nrowse = div_v*((nrows+div_v-1)/div_v);
  *ncolse = div_h*((ncols+div_h-1)/div_h);
}


int least_common_multiplier(int a, int b)
{
  int n;
  
  if (a<1 || b<1) 
	 return 0;  /*Error*/

  n=max(a,b);

  while(n%a != 0 || n%b != 0)
	 n++;
	
  return n;
}

/***************************************************************************
* Function for writing one frame to file                 
*--------------------------------------------------------------------------
* Input:     nrowst  - total number of rows
*            ncolst  - total number of columns
*            top     - first row to read
*            left    - frist column to read
*            nrows   - number of rows to read
*            ncols   - number of columns to read
*            depth   - bit depth
*            x       - frame                                   
*            fp      - file pointer
*--------------------------------------------------------------------------
* First version:          Arild Fuldseth        1996-01-31          
***************************************************************************/
 
void Putframe(int nrowst,int ncolst,int top,int left,int nrows,int ncols,
	      int depth,float *x, FILE *fp){
 
  int i,j,nbytes,col,row;
  float tmp,mean;

  nbytes = (depth+7)/8;
  mean = (depth > 1) ? (float)(1<<(depth-1)) : 0.0;
 

  i = 0;
  fseek(fp,(nbytes*ncolst*top),SEEK_SET);
  for (row=0;row<nrows;row++){
    fseek(fp,(nbytes*left),SEEK_CUR);
    for (col=0;col<ncols;col++){
      tmp = x[i++] + mean;
		if (tmp < 0)
		  tmp = 0;
		if (tmp > (1<<depth)-1)
		  tmp = (1<<depth)-1;
      for (j=8*(nbytes-1);j>=0;j-=8) fputc((char)((long)tmp>>j),fp);
    }
    fseek(fp,(nbytes*(ncolst-ncols-left)),SEEK_CUR);
  }
}

void Putframe2(int N,float *x, FILE *fp){
 
  int i;
 
  for (i=0;i<N;i++) {
  	
    if (x[i] > 255)
  		x[i] = 255;
  	else if (x[i] < 0)
  		x[i] = 0;
  	fputc((unsigned char)((int)x[i]),fp);
  	
  }
}
 
/***************************************************************************
* Function for reading one frame from file                 
*--------------------------------------------------------------------------
* Input:     nrowst  - total number of rows
*            ncolst  - total number of columns
*            top     - first row to read
*            left    - frist column to read
*            nrows   - number of rows to read
*            ncols   - number of columns to read
*            depth   - bit depth 
*            fp      - file pointer
* Output:    x       - frame                                   
*--------------------------------------------------------------------------
* First version:          Arild Fuldseth        1996-01-31          
***************************************************************************/
 
void Getframe(int nrowst,int ncolst,int top,int left,int nrows,int ncols,
	      int depth,float *x, FILE *fp){
 
  int i,j,row,col,nbytes;
  float mean;

  nbytes = (depth+7)/8;
  mean = (depth > 1) ? (float)(1<<(depth-1)) : 0.0; /* 2^(depth-1) */

  i = 0;
  fseek(fp,(nbytes*ncolst*top),SEEK_SET);
  for (row=0;row<nrows;row++){
    fseek(fp,(nbytes*left),SEEK_CUR);
    for (col=0;col<ncols;col++){
      x[i] = -mean;
      for (j=8*(nbytes-1);j>=0;j-=8) x[i] += (float)((long)fgetc(fp) << j);
      if (i >= nrows*ncols) printf("%4d %4d %4d\n",row,col,i);
      i++;
    }
    fseek(fp,(nbytes*(ncolst-ncols-left)),SEEK_CUR);
  }
}
 
/***************************************************************************
* Function for extension of image  
*--------------------------------------------------------------------------
* Input:   nrows     - number of rows in frame
*          ncols     - number of columns in frame 
*          nrowse    - extended number of rows in frame
*          ncolse    - extended number of columns in frame 
*          x         - original image
*          qflag     - quiet flag  
* Output:  xe        - extended image
*--------------------------------------------------------------------------
* First version:  Arild Fuldseth 
* Last version:   Helge Coward            2001-03-08
***************************************************************************/
 
void Extend_Image(int nrows,int ncols,int nrowse, int ncolse, 
						float	*x,float *xe, int screenflag){ 
 
  int i,j,nrows1,nrows2,ncols1,ncols2;
  float *fptr1,*fptr2;

  nrows1 = (nrowse-nrows)/2;
  nrows2 = nrowse-nrows-nrows1;

  ncols1 = (ncolse-ncols)/2;
  ncols2 = ncolse-ncols-ncols1;
 
  if (screenflag)
	 fprintf(stderr,"Expanding image from %1d X %1d to %1d X %1d\n",
			  nrows,ncols,nrowse,ncolse);

  fptr2 = xe;
  for (i=0;i<nrows1;i++){
    fptr1 = &x[0];
    for (j=0;j<ncols1;j++) *fptr2++ = *fptr1; 
    for (j=0;j<(ncols-1);j++) *fptr2++ = *fptr1++;
    for (j=0;j<(ncols2+1);j++) *fptr2++ = *fptr1; 
  }
  for (i=0;i<nrows;i++){
     fptr1 = &x[i*ncols];
     for (j=0;j<ncols1;j++) *fptr2++ = *fptr1;
     for (j=0;j<(ncols-1);j++) *fptr2++ = *fptr1++;
     for (j=0;j<(ncols2+1);j++) *fptr2++ = *fptr1; 
  }
  for (i=0;i<nrows2;i++){
     fptr1 = &x[(nrows-1)*ncols];
     for (j=0;j<ncols1;j++) *fptr2++ = *fptr1;
     for (j=0;j<(ncols-1);j++) *fptr2++ = *fptr1++;
     for (j=0;j<(ncols2+1);j++) *fptr2++ = *fptr1; 
   }
}

/***************************************************************************
* Function for reduction of image  
*--------------------------------------------------------------------------
* Input:   nrows     - number of rows in frame
*          ncols     - number of columns in frame 
*          nrowse    - extended number of rows in frame
*          ncolse    - extended number of columns in frame 
*          xe        - extended image
*          qflag     - quiet flag  
* Output:  x         - reduced image
*--------------------------------------------------------------------------
* First version:  Arild Fuldseth 
* Last version:   Helge Coward            2001-03-08
***************************************************************************/
 
void Reduce_Image(int nrows, int ncols, int nrowse, int ncolse, 
						float	*xe,float *x, int screenflag)
{ 
  int i,j,nrows1,nrows2,ncols1,ncols2;
  float *fptr1,*fptr2;

  nrows1 = (nrowse-nrows)/2;
  nrows2 = nrowse-nrows-nrows1;

  ncols1 = (ncolse-ncols)/2;
  ncols2 = ncolse-ncols-ncols1;
 
  if (screenflag)
	 fprintf(stderr,"Reducing image from %1d X %1d to %1d X %1d\n",
			  nrowse,ncolse,nrows,ncols);

  fptr2 = xe;
  for (i=0;i<nrows1;i++){
    fptr1 = &x[0];
    for (j=0;j<ncols1;j++) *fptr1 = *fptr2++; 
    for (j=0;j<(ncols-1);j++) *fptr1++ = *fptr2++;
    for (j=0;j<(ncols2+1);j++) *fptr1 = *fptr2++; 
  }
  for (i=0;i<nrows;i++){
     fptr1 = &x[i*ncols];
     for (j=0;j<ncols1;j++) *fptr1 = *fptr2++;
     for (j=0;j<(ncols-1);j++) *fptr1++ = *fptr2++;
     for (j=0;j<(ncols2+1);j++) *fptr1 = *fptr2++; 
  }
  for (i=0;i<nrows2;i++){
     fptr1 = &x[(nrows-1)*ncols];
     for (j=0;j<ncols1;j++) *fptr1 = *fptr2++;
     for (j=0;j<(ncols-1);j++) *fptr1++ = *fptr2++;
     for (j=0;j<(ncols2+1);j++) *fptr1 = *fptr2++; 
   }
}


void dct(float **input,float **output, int nrows, int ncols,int buffer) {
	int i,k,n;	
	double Pi=3.141592654,sum,w[buffer],cosvalue;
	w[0] = sqrt(1.0/buffer);
	for (i=1;i<buffer;i++) {
		w[i] = sqrt(2.0/buffer);
	}
	//printf("w[0] = %lf og w[1] = %lf \n",w[0],w[1]);	
	
	for (i=0;i<nrows*ncols;i++) {
		for (k=0;k<buffer;k++) {	
			sum=0.0;
			for (n=0; n<buffer; n++) {
				cosvalue = cos((Pi*(2*n+1)*k)/(2*buffer));
				sum += cosvalue*(double)input[i][n];			
			}
			output[i][k] = (float)(w[k]*sum);
			//printf("w[k] = %lf og sum= %lf og output[i][k]= %f \n",w[k],sum,output[i][k]);
		}
	}	
}

void idct(float **input,float **output, int nrows, int ncols,int buffer) {
	int i,k,n;	
	double Pi=3.141592654,sum,w[2],cosvalue;
	w[0] = sqrt(1.0/buffer);
	w[1] = sqrt(2.0/buffer);
	
	for (i=0;i<nrows*ncols;i++) {
		for (k=0;k<buffer;k++) {	
			sum=0.0;
			for (n=1; n<buffer; n++) {
				cosvalue = cos((2*k+1)*Pi*n/(2*buffer));
				sum += cosvalue*(double)input[i][n];			
			}
			output[i][k] = (float)(w[0]*input[i][0])+(float)(w[1]*sum);
			//printf("w[k] = %lf og sum= %lf og output[i][k]= %f \n",w[k],sum,output[i][k]);
		}
	}	
}


void Getframe2(int N, int Framenr, float **x, FILE *fp){

  int i;

  for (i=0;i<N;i++) x[i][Framenr] = (float)fgetc(fp);
  //for (i=0;i<N;i++) x[i][Framenr] = (float)fgetc(fp)-128.0;
}


void writeTxtFile1_f(char txtfile[80],float *input,int length) {
	FILE *ftxt;
	int i;
	if ((ftxt = fopen(txtfile,"wr")) == NULL){
    	fprintf(stderr,"Cannot open file %s for writing\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<length;i++) {
  		fprintf(ftxt, "%f", input[i]);
  		if (i != (length-1)){
  			fprintf(ftxt, " ");
  		}
  	}
 	//close file
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);	
  	}	
}

void writeTxtFile1_i(char txtfile[80],int *input,int length) {
	FILE *ftxt;
	int i;
	if ((ftxt = fopen(txtfile,"wr")) == NULL){
    	fprintf(stderr,"Cannot open file %s for writing\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<length;i++) {
  		fprintf(ftxt, "%i", input[i]);
  		if (i!= (length-1)) {
  			fprintf(ftxt, " ");
  		}
  	}
 	//close file
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);	
  	}	
}

//skriver ut unsigned char matrise til text-fil
void writeTxtFile2_uch(char txtfile[80],unsigned char **input,int nrows,int ncols) {
	FILE *ftxt;
	int i,j;
	if ((ftxt = fopen(txtfile,"wr")) == NULL){
    	fprintf(stderr,"Cannot open file %s for writing\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<nrows;i++) {
  		for (j=0;j<ncols;j++) {  			  		
	  		fprintf(ftxt, "%i", (int)input[i][j]);
	  		if (j!=(ncols-1)) {
	  			fprintf(ftxt, " ");
	  		}
	  		else if (j==(ncols-1) && i!=(nrows-1)) {
	  			fprintf(ftxt, "\n");
	  		}
  		}
  	}
 	//close file
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);	
  	}	
}

void writeTxtFile2_f(char txtfile[80],float **input,int nrows,int ncols) {
	FILE *ftxt;
	int i,j;
	if ((ftxt = fopen(txtfile,"wr")) == NULL){
    	fprintf(stderr,"Cannot open file %s for writing\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<nrows;i++) {
  		for (j=0;j<ncols;j++) {  		
	  		fprintf(ftxt, "%f", input[i][j]);
	  		if (j!=(ncols-1)) {
	  			fprintf(ftxt, " ");
	  		}
	  		else if (j==(ncols-1) && i!=(nrows-1)) {
	  			fprintf(ftxt, "\n");
	  		}
  		}
  	}
 	//close file
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);	
  	}	
}

void writeTxtFile2_si(char txtfile[80],short int **input,int nrows,int ncols) {
	FILE *ftxt;
	int i,j;
	if ((ftxt = fopen(txtfile,"wr")) == NULL){
    	fprintf(stderr,"Cannot open file %s for writing\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<nrows;i++) {
  		for (j=0;j<ncols;j++) {  		
	  		fprintf(ftxt, "%i", input[i][j]);
	  		if (j!=(ncols-1)) {
	  			fprintf(ftxt, " ");
	  		}
	  		else if (j==(ncols-1) && i!=(nrows-1)) {
	  			fprintf(ftxt, "\n");
	  		}
  		}
  	}
 	//close file
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);	
  	}	
}


void writeTxtFile2_d(char txtfile[80],double input[4][256],int nrows,int ncols) {
	FILE *ftxt;
	int i,j;
	if ((ftxt = fopen(txtfile,"wr")) == NULL){
    	fprintf(stderr,"Cannot open file %s for writing\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<nrows;i++) {
  		for (j=0;j<ncols;j++) {  		
	  		fprintf(ftxt, "%lf", input[i][j]);
	  		if (j!=(ncols-1)) {
	  			fprintf(ftxt, " ");
	  		}
	  		else if (j==(ncols-1) && i!=(nrows-1)) {
	  			fprintf(ftxt, "\n");
	  		}
  		}
  	}
 	//close file
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);	
  	}	
}

void writeTxtFile2_uss(char txtfile[80],unsigned short input[6][256],int nrows,int ncols) {
	FILE *ftxt;
	int i,j;
	if ((ftxt = fopen(txtfile,"wr")) == NULL){
    	fprintf(stderr,"Cannot open file %s for writing\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<nrows;i++) {
  		for (j=0;j<ncols;j++) {  		
	  		fprintf(ftxt, "%u", input[i][j]);
	  		if (j!=(ncols-1)) {
	  			fprintf(ftxt, " ");
	  		}
	  		else if (j==(ncols-1) && i!=(nrows-1)) {
	  			fprintf(ftxt, "\n");
	  		}
  		}
  	}
 	//close file
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);	
  	}	
}

void writeTxtFile1_uch(char txtfile[80],unsigned char *input,int length) {
	FILE *ftxt;
	int i;
	if ((ftxt = fopen(txtfile,"wr")) == NULL){
    	fprintf(stderr,"Cannot open file %s for writing\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<length;i++) {
	  	fprintf(ftxt, "%i", (int)input[i]);
	  	if (i!=(length-1)) {
	  		fprintf(ftxt, " ");
	  	}	  	  		
  	}
 	//close file
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);	
  	}	
}

void writeTxtFile1_i_table(char txtfile[80],int Input[MAXCBITS/2],long length) {
	FILE *ftxt;
	int i;
	if ((ftxt = fopen(txtfile,"wr")) == NULL){
    	fprintf(stderr,"Cannot open file %s for writing\n",txtfile);
	    //exit(1);
	    return;
  	}
  	for (i=0;i<length;i++) {
	  	fprintf(ftxt, "%i", (int)Input[i]);
	  	if (i!=(length-1)) {
	  		fprintf(ftxt, " ");
	  	}	  	  		
  	}
 	//close file
  	i = fclose(ftxt);
  	if (i != 0) {
  		fprintf(stderr, "fclose returned %i \n",i);	
  	}	
}
