#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include "codecdecl.h"

/*!
 ************************************************************************
 * \brief
 *    init the Dekoder CodingEnvironment used for CABAC encoding
 ************************************************************************
 */
 
CABAC_DecoderEnvironment* init_decoder_environment(unsigned char **Code, int *CodeCnt) {
	CABAC_DecoderEnvironment* env;
	
	env = (CABAC_DecoderEnvironment *)malloc(sizeof(CABAC_DecoderEnvironment));
	
	env->mot_ctx = create_contexts_MotionInfo();
	env->tex_ctx = create_contexts_TextureInfo();
	env->partArr = create_decoder_dataPartition(Code,CodeCnt);
	return env;
}

/*!
 ************************************************************************
 * \brief
 *    init the Encoder CodingEnvironment used for CABAC encoding
 ************************************************************************
 */
 
CABAC_EncoderEnvironment* init_encoder_environment(unsigned char **Code, int *CodeCnt) {
	CABAC_EncoderEnvironment* env;
	
	env = (CABAC_EncoderEnvironment *)malloc(sizeof(CABAC_EncoderEnvironment));
	
	env->mot_ctx = create_contexts_MotionInfo();
	env->tex_ctx = create_contexts_TextureInfo();
	env->partArr = create_encoder_dataPartition(Code,CodeCnt);
	return env;
}


void delete_decoder_environment(CABAC_DecoderEnvironment* eep) {
	delete_contexts_MotionInfo(eep->mot_ctx);
	delete_contexts_TextureInfo(eep->tex_ctx);
	delete_decoder_DataPartition(eep->partArr);
	free(eep);
}

void delete_encoder_environment(CABAC_EncoderEnvironment* eep) {
	delete_contexts_MotionInfo(eep->mot_ctx);
	delete_contexts_TextureInfo(eep->tex_ctx);
	delete_encoder_DataPartition(eep->partArr);
	free(eep);
}
	

/*!
 ************************************************************************
 * \brief
 *    Allocation of contexts models for the motion info
 *    used for arithmetic encoding
 ************************************************************************
 */
MotionInfoContexts* create_contexts_MotionInfo(void)
{
  MotionInfoContexts* enco_ctx;

  enco_ctx = (MotionInfoContexts*) calloc(1, sizeof(MotionInfoContexts) );
  if( enco_ctx == NULL ) {
    fprintf(stderr,"error: create_contexts_MotionInfo: enco_ctx");
    exit(ERROR);
    //no_mem_exit("create_contexts_MotionInfo: enco_ctx");
  }

  return enco_ctx;
}


/*!
 ************************************************************************
 * \brief
 *    Allocates of contexts models for the texture info
 *    used for arithmetic encoding
 ************************************************************************
 */
TextureInfoContexts* create_contexts_TextureInfo(void)
{
  TextureInfoContexts*  enco_ctx;

  enco_ctx = (TextureInfoContexts*) calloc(1, sizeof(TextureInfoContexts) );
  if( enco_ctx == NULL ) {
    fprintf(stderr, "error: create_contexts_TextureInfo: enco_ctx");
    exit(ERROR);
    //no_mem_exit("create_contexts_TextureInfo: enco_ctx");
  }

  return enco_ctx;
}

/*!
 ************************************************************************
 * \brief
 *    Allocates of DataPartition
 *    used for arithmetic encoding
 ************************************************************************
 */
 
Decoder_DataPartition* create_decoder_dataPartition(unsigned char **Code, int *CodeCnt) {
 	Decoder_DataPartition* d_p;
 	
 	d_p = (Decoder_DataPartition*) calloc(1,sizeof(Decoder_DataPartition));
 	if (d_p == NULL) {
 		fprintf(stderr, "error: create_DataPartition: d_p");
    	exit(ERROR);   
  	}
  	
  	if ((d_p->bitstream = (Bitstream*) calloc(1, sizeof(Bitstream)))==NULL) {
  		fprintf(stderr, "error: create_DataPartition: d_p->bitstream");
  		exit(ERROR);
  	}  	
  	d_p->bitstream->streamBuffer = Code[0];
  	d_p->bitstream->byte_pos = *CodeCnt;  	  	
  	return d_p; 	 	
}


/*!
 ************************************************************************
 * \brief
 *    Allocates of DataPartition
 *    used for arithmetic encoding
 ************************************************************************
 */
 
Encoder_DataPartition* create_encoder_dataPartition(unsigned char **Code, int *CodeCnt) {
 	Encoder_DataPartition* d_p;
 	
 	d_p = (Encoder_DataPartition*) calloc(1,sizeof(Encoder_DataPartition));
 	if (d_p == NULL) {
 		fprintf(stderr, "error: create_encoder_DataPartition: d_p");
    	exit(ERROR);   
  	}
  	
  	if ((d_p->bitstream = (Bitstream*) calloc(1, sizeof(Bitstream)))==NULL) {
  		fprintf(stderr, "error: create_encoder_DataPartition: d_p->bitstream");
  		exit(ERROR);
  	}  	
  	d_p->bitstream->streamBuffer = Code[0];
  	d_p->bitstream->byte_pos = *CodeCnt;  	  	
  	return d_p; 	 	
}


/*!
 ************************************************************************
 * \brief
 *    Frees the memory of the contexts models
 *    used for arithmetic encoding of the motion info.
 ************************************************************************
 */
void delete_contexts_MotionInfo(MotionInfoContexts *enco_ctx)
{
  if( enco_ctx == NULL )
    return;

  free( enco_ctx );

  return;
}

/*!
 ************************************************************************
 * \brief
 *    Frees the memory of the contexts models
 *    used for arithmetic encoding of the texture info.
 ************************************************************************
 */
void delete_contexts_TextureInfo(TextureInfoContexts *enco_ctx)
{
  if( enco_ctx == NULL )
    return;

  free( enco_ctx );

  return;
}

/*!
 ************************************************************************
 * \brief
 *    Frees the memory of the DataPartition
 *    used for arithmetic encoding of the texture info.
 ************************************************************************
 */
void delete_decoder_DataPartition(Decoder_DataPartition *d_p)
{
  if( d_p == NULL )
    return;
  
  if( d_p->bitstream != NULL)
  	free(d_p->bitstream);
  
  	
  free( d_p );

  return;
}

/*!
 ************************************************************************
 * \brief
 *    Frees the memory of the DataPartition
 *    used for arithmetic encoding of the texture info.
 ************************************************************************
 */
void delete_encoder_DataPartition(Encoder_DataPartition *d_p)
{
  if( d_p == NULL )
    return;
  
  if( d_p->bitstream != NULL)
  	free(d_p->bitstream);
  
  	
  free( d_p );

  return;
}


